#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ido.h"


#define  blockDim_x       16
#define  blockDim_y       16


__global__  void  cuda_diffusion2d_0
// ====================================================================
//
// program    :  CUDA device code for 2-D diffusion equation
//               for 16 x 16 block and 16 x 16 thread per 1 block
//
// date       :  Nov 07, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *f,         /* dependent variable                        */
   float    *fn,        /* dependent variable                        */
   int      nx,         /* grid number in the x-direction            */
   int      ny,         /* grid number in the x-direction            */
   float    c0,         /* coefficient no.0                          */
   float    c1,         /* coefficient no.1                          */
   float    c2          /* coefficient no.2                          */
)
// --------------------------------------------------------------------
{
   int    j,    jx,   jy;
   float  fcc,  fce,  fcw,  fcs,  fcn;

   jy = blockDim.y*blockIdx.y + threadIdx.y;
   jx = blockDim.x*blockIdx.x + threadIdx.x;
   j = nx*jy + jx;

   fcc = f[j];

   if(jx == 0) fcw = fcc;
   else        fcw = f[j - 1];

   if(jx == nx - 1) fce = fcc;
   else             fce = f[j+1];

   if(jy == 0) fcs = fcc;
   else        fcs = f[j-nx];

   if(jy == ny - 1) fcn = fcc;
   else             fcn = f[j+nx];

   fn[j] = c0*(fce + fcw)
         + c1*(fcn + fcs)
         + c2*fcc;
}



__global__  void  cuda_diffusion2d_1
// ====================================================================
//
// program    :  CUDA device code for 2-D diffusion equation
//               for 16 x 16 block and 16 x 16 thread per 1 block
//
// date       :  May 13, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *f,         /* dependent variable                        */
   float    *fn,        /* dependent variable                        */
   int      nx,         /* grid number in the x-direction            */
   int      ny,         /* grid number in the x-direction            */
   float    c0,         /* coefficient no.0                          */
   float    c1,         /* coefficient no.1                          */
   float    c2          /* coefficient no.2                          */
)
// --------------------------------------------------------------------
{
        int   j,   jx = threadIdx.x + 1,  jy = threadIdx.y + 1;
        float fcc;

       __shared__   float  fs[blockDim_y+2][blockDim_x+2];

       j = nx*(blockDim.y*blockIdx.y + threadIdx.y)
         + blockDim.x*blockIdx.x + threadIdx.x;

       fcc = f[j];
       fs[jy][jx] = fcc;


       if(threadIdx.x == 0) {
          if(blockIdx.x == 0) fs[jy][0] = fcc;
          else                fs[jy][0] = f[j-1];
       }

       if(threadIdx.x == blockDim.x - 1) {
          if(blockIdx.x == gridDim.x - 1) fs[jy][blockDim.x + 1] = fcc;
          else                            fs[jy][blockDim.x + 1] = f[j+1];
       }

       if(threadIdx.y ==  0) {
          if(blockIdx.y == 0) fs[0][jx] = fcc;
          else                fs[0][jx] = f[j - nx];
       }

       if(threadIdx.y ==  blockDim.y - 1) {
          if(blockIdx.y == gridDim.y - 1) fs[blockDim.y + 1][jx] = fcc;
          else                            fs[blockDim.y + 1][jx] = f[j + nx];
       }


       __syncthreads();

       fn[j] = c0*(fs[jy][jx+1] + fs[jy][jx-1])
             + c1*(fs[jy+1][jx] + fs[jy-1][jx])
             + c2*fs[jy][jx];
}



void  cpu_diffusion2d
// ====================================================================
//
// purpos     :  2-dimensional diffusion equation solved by FDM
//
// date       :  May 16, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *f,         /* dependent variable                        */
   float    *fn,        /* updated dependent variable                */
   int      nx,         /* x-dimensional grid size                   */
   int      ny,         /* y-dimensional grid size                   */
   float    c0,         /* coefficient no.0                          */
   float    c1,         /* coefficient no.1                          */
   float    c2          /* coefficient no.2                          */
)
// --------------------------------------------------------------------
{
     int    j,    jx,   jy;
     float  fcc,  fce,  fcw,  fcs,  fcn;

     for(jy = 0; jy < ny; jy++) {
         for(jx = 0; jx < nx; jx++) {
             j = nx*jy + jx;
             fcc = f[j];

             if(jx == 0) fcw = fcc;
             else        fcw = f[j - 1];

             if(jx == nx - 1) fce = fcc;
             else             fce = f[j+1];

             if(jy == 0) fcs = fcc;
             else        fcs = f[j-nx];

             if(jy == ny - 1) fcn = fcc;
             else             fcn = f[j+nx];

             fn[j] = c0*(fce + fcw)
                   + c1*(fcn + fcs)
                   + c2*fcc;
         }
     }
}



float  diffusion2d
// ====================================================================
//
// purpos     :  2-dimensional diffusion equation solved by FDM
//
// date       :  May 16, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   int      numGPUs,    /* available number of GPU device            */
   int      nx,         /* x-dimensional grid size                   */
   int      ny,         /* y-dimensional grid size                   */
   float    *f,         /* dependent variable                        */
   float    *fn,        /* updated dependent variable                */
   float    kappa,      /* diffusion coefficient                     */
   float    dt,         /* time step interval                        */
   float    dx,         /* grid spacing in the x-direction           */
   float    dy          /* grid spacing in the y-direction           */
)
// --------------------------------------------------------------------
{
     float     c0 = kappa*dt/(dx*dx),   c1 = kappa*dt/(dy*dy),
               c2 = 1.0 - 2.0*(c0 + c1);

     if(numGPUs > 0) {

        dim3  grid(nx/blockDim_x,ny/blockDim_y,1),  threads(blockDim_x,blockDim_y,1);
        cuda_diffusion2d_0<<< grid, threads >>>(f,fn,nx,ny,c0,c1,c2);

        cudaThreadSynchronize();
     }
     else cpu_diffusion2d(f,fn,nx,ny,c0,c1,c2);

     return (float)(nx*ny)*7.0;
}
